import unittest

from cad_parser.definitions import Order, Patient, Element

class TestOrder(unittest.TestCase):
    def test_constructor(self):
        order = Order(
            project='foo', 
            format='exocad', 
            patient=Patient(1, 'Mustermann', 'Max'), 
            comment=':(){ :|:& };:', 
            practice_id=1, 
            elements=[
                Element(file_name='foo.stl', material='gold', color='A1', height=7.5, is_screw_retained=None, type='ModelBase', teeth=[])
            ],
            models=[],
            scan_files=[],
            loaded_scene_file=None
        )

        self.assertEqual(order.project, 'foo')
        self.assertEqual(order.format, 'exocad')
        self.assertEqual(order.patient.id, 1)
        self.assertEqual(order.patient.last_name, 'Mustermann')
        self.assertEqual(order.patient.first_name, 'Max')
        self.assertEqual(order.comment, ':(){ :|:& };:')
        self.assertEqual(order.practice_id, 1)
        self.assertEqual(order.elements[0].file_name, 'foo.stl')
        self.assertEqual(order.elements[0].material, 'gold')
        self.assertEqual(order.elements[0].color, 'A1')
        self.assertEqual(order.elements[0].height, 7.5)
        self.assertEqual(order.elements[0].is_screw_retained, None)
        self.assertEqual(order.elements[0].teeth, [])



import unittest

from cad_parser.definitions import Patient

class TestPatient(unittest.TestCase):
    def test_constructor(self):
        patient = Patient(1, 'Mustermann', 'Max')

        self.assertEqual(patient.id, 1)
        self.assertEqual(patient.last_name, 'Mustermann')
        self.assertEqual(patient.first_name, 'Max')
        


import unittest

from cad_parser.definitions import Element, Tooth

class TestElement(unittest.TestCase):
    def test_constructor(self):
        teeth = [
            Tooth(fdi=1, color='A1', implant_system=None, implant_type=None, is_anatomical=None, type='AnatomicCrown'),
            Tooth(fdi=2, color='A2', implant_system=None, implant_type=None, is_anatomical=None, type='AnatomicCrown')
        ]
        element = Element(file_name='foo.stl', material='gold', color='orange', height=7.838345409284646, is_screw_retained=None, type=None, teeth=teeth)

        self.assertEqual(element.file_name, 'foo.stl')
        self.assertEqual(element.material, 'gold')
        self.assertEqual(element.color, 'orange')
        self.assertAlmostEqual(element.height, 7.838345409284646)
        self.assertEqual(element.is_screw_retained, None)
        
        self.assertEqual(len(element.teeth), 2)

        tooth0 = element.teeth[0]
        self.assertEqual(tooth0.fdi, 1)
        self.assertEqual(tooth0.color, 'A1')
        self.assertEqual(tooth0.implant_type, None)
        self.assertEqual(tooth0.is_anatomical, None)
        self.assertEqual(tooth0.type, 'AnatomicCrown')

        tooth1 = element.teeth[1]
        self.assertEqual(tooth1.fdi, 2)
        self.assertEqual(tooth1.color, 'A2')
        self.assertEqual(tooth1.implant_type, None)
        self.assertEqual(tooth1.is_anatomical, None)
        self.assertEqual(tooth1.type, 'AnatomicCrown')



import unittest

from cad_parser import exocad

from pathlib import Path
import logging

class SkipTest:
    def __init__(self, condition, description):
        self.condition = condition
        self.description = description

    def __call__(self, func):
        if not self.condition:
            return unittest.skip(self.description)(func)
        return func

@SkipTest(Path('tests/samples/exocad').exists(), "Lokal tests only")
class TestExocad(unittest.TestCase):
    logger = logging.getLogger(__name__)
    logging.basicConfig(format = '%(asctime)s %(module)s %(levelname)s: %(message)s', datefmt = '%m/%d/%Y %I:%M:%S %p', level = logging.INFO)

    def test_is_supported(self):
        project_types = Path('tests/samples/exocad').iterdir()
        project_types = filter(lambda x: x.is_dir(), project_types)
        for project_type in project_types:
            for project in project_type.iterdir():
                if project.name == 'MissingConstructionInfo' or project.is_file():
                    continue
                
                self.assertTrue(exocad.is_supported(project), f"{project} is not supported")
                
    def test_parse_dental_project(self):
        project_types = Path('tests/samples/exocad').iterdir()
        project_types = filter(lambda x: x.is_dir(), project_types)
        for project_type in project_types:
            for project in project_type.iterdir():
                if project.name == 'MissingConstructionInfo' or project.is_file():
                    continue
                dental_project_path = next(project.glob('*.dental[pP]roject'))
                patient = exocad.parse_dental_project(dental_project_path)
                self.assertIsNotNone(patient.id)
                self.assertIsNotNone(patient.last_name)

    def test_parse_construction_info(self):
        project_types = Path('tests/samples/exocad').iterdir()
        project_types = filter(lambda x: x.is_dir(), project_types)
        for project_type in project_types:
            for project in project_type.iterdir():
                construction_info_path = next(project.glob('*.constructionInfo'), None)
                if construction_info_path:
                    order = exocad.parse_construction_info(construction_info_path)

    def test_parse_model_info(self):
        project_types = Path('tests/samples/exocad').iterdir()
        project_types = filter(lambda x: x.is_dir(), project_types)
        for project_type in project_types:
            for project in project_type.iterdir():
                model_info_path = next(project.glob('*.modelInfo'), None)
                if model_info_path:
                    order = exocad.parse_model_info(model_info_path)
                    
    def test_parse_scan_info(self):
        project_types = Path('tests/samples/exocad').iterdir()
        project_types = filter(lambda x: x.is_dir(), project_types)
        for project_type in project_types:
            for project in project_type.iterdir():
                scan_info_path = next(project.glob('*.scanInfo'), None)
                if scan_info_path:
                    order = exocad.parse_scan_info(scan_info_path)

    def test_parse_partial_info(self):
        project_types = Path('tests/samples/exocad').iterdir()
        project_types = filter(lambda x: x.is_dir(), project_types)
        for project_type in project_types:
            for project in project_type.iterdir():
                partial_info_path = next(project.glob('*.partialInfo'), None)
                if partial_info_path:
                    order = exocad.parse_partial_info(partial_info_path)
    
    def test_parse(self):
        project_types = Path('tests/samples/exocad').iterdir()
        project_types = filter(lambda x: x.is_dir(), project_types)
        for project_type in project_types:
            for project in project_type.iterdir():
                if project.name == 'MissingConstructionInfo' or project.is_file():
                    continue
                order = exocad.parse(project)

    def test_parse_partial_clamp(self):
        project = Path('tests/samples/exocad/32_Modellguss/00001_Viktor_Klammer_MG_-2025-04-03_010/Viktor_Klammer_MG_-2025-04-03-lowerjaw.partialInfo')
        elements = exocad.parse_partial_info(project)
        for element in elements:
            self.assertEqual(element.type, 'modelclip')

    def test_parse_partial_framework(self):
        project = Path('tests/samples/exocad/32_Modellguss/00001_Viktor_Tertiaer_-2025-04-02_010/Viktor_Tertiaer_-2025-04-02-upperjaw.partialInfo')
        elements = exocad.parse_partial_info(project)
        for element in elements:
            self.assertEqual(element.type, 'framework')

    def test_parse_construction_info(self):
        project = Path('tests/samples/exocad/32_Modellguss/00001_Viktor_Tertiaer_-2025-04-03_008_aus_CAD/Viktor_Tertiaer_-2025-04-03.constructionInfo')
        order = exocad.parse_construction_info(project)
        self.logger.info(order)
        for element in order.elements:
            self.assertEqual(element.type,  'MergedBridgeSLM')

    def test_construction_info_with_model_info(self):
        project = Path('tests/samples/exocad/33_Construction+Modell/03369_Luedke_Ralf_-2025-04-11_001')
        _, _, _, _, _, order = exocad.parse(project)
        self.logger.info(order.models)
        self.assertGreater(len(order.elements), 0)
        self.assertGreater(len(order.models), 0)

    def test_abutment_type(self):
        project = Path('tests/samples/exocad/09_Abutment/154711')
        _, _, _, _, _, order = exocad.parse(project)
        self.assertEqual(order.elements[0].teeth[0].implant_system, 'medentika_preface-t_conid : PreFace_T9000-R : T9000-R')

        project = Path('tests/samples/exocad/09_Abutment/156094_krone_inkl_Abutment')
        _, _, _, _, _, order = exocad.parse(project)
        self.assertEqual(order.elements[0].teeth[0].implant_system, 'medentika_preface-l_conid : PreFace_L9810-16R-1 : L9810-16R-1')
        self.assertEqual(order.elements[1].teeth[0].implant_system, None)

        project = Path('tests/samples/exocad/EdgeCases/AbutmentAliasNotDetected')
        _, _, _, _, _, order = exocad.parse(project)
        self.assertEqual(order.elements[0].teeth[0].implant_system, 'medentika_preface-n_conid : PreFace_N9810-16R-1 : N9810-16R-1')

    def test_parse_complex_partial_framework(self):
        project = Path('tests/samples/exocad/EdgeCases/PartialFramework')
        found_dentalProject = next(project.glob('*.dental[pP]roject'), False)
        found_constructionInfo = next(project.glob('*.construction[iI]nfo'), False)
        found_modelInfo = next(project.glob('*.model[iI]nfo'), False)
        found_scanInfo = next(project.glob('*.scan[iI]nfo'), False)
        found_partialInfo = next(project.glob('*.partial[iI]nfo'), False)

        print('found_dentalProject:', found_dentalProject)
        print('found_constructionInfo:', found_constructionInfo)
        print('found_modelInfo:', found_modelInfo)
        print('found_scanInfo:', found_scanInfo)
        print('found_partialInfo:', found_partialInfo)



import io

import lxml.etree

from cad_parser.errors import CadParserError

NONE_VALUE_STRINGS = ['', 'None', 'Keine', '-']

def consolidate_colors(teeth):
    colors = list(set([t.color for t in teeth if t.color]))

    if len(colors) == 0:
        return None

    if len(colors) > 1:
        return 'Mixed'

    return colors[0]

def get_text(node, xml_path: str, ignore_nullable: bool=True):
    target = node.find(xml_path)
    if target == None: return
    if ignore_nullable and (target.text in NONE_VALUE_STRINGS): return
    return target.text

def parse_document(source: str | io.IOBase):
    if isinstance(source, str):
        return lxml.etree.fromstring(source)
    elif isinstance(source, io.IOBase):
        return lxml.etree.parse(source)
    else:
        raise NotImplementedError(f'{source} is neither str nor a file-like object')

def unn_to_fdi(number: int) -> int:
    if number >= 1 and number <= 8:
        return 19 - number
    if number >= 9 and number <= 16:
        return 12 + number
    if number >= 17 and number <= 24:
        return 55 - number
    if number >= 25 and number <= 32:
        return 16 + number

    raise CadParserError(f'Invalid tooth indication number {number}')



from typing import Tuple, List
from pathlib import Path

from .errors import HyperdentParserError
from .definitions import Order, Element, Tooth
from .utils import parse_document, get_text

FORMAT_NAME = 'hyperdent'

def is_supported(file_path: Path) -> bool:
    '''
    Is the given file path a supported hyperdent project?

    Parameters
    ----------
    file_path : Path
        The file path to check.

    Returns
    -------
    bool
        True if the file path is a supported hyperdent project, False otherwise.
    '''
    return next(file_path.glob('*.hdpp'), False)

def parse(project_path: Path) -> Tuple[List[Path], Order]:
    '''
    Parse a project consisting of hdpp files and return an `Order` object.

    Parameters
    ----------
    project_path : Path
        The path to the exocad project directory.

    Returns
    -------
    files : [Path]
        A list of all hdpp files within the folder.
    order : Order
        The order object containing the elements defined in the hdpp files.
    '''
    if not is_supported(project_path):
        raise HyperdentParserError(f'Unsupported hyperdent project: {project_path}')
    
    hdpp_files = project_path.glob('*.hdpp')
    files = []
    elements = []
    while hdp_file := next(hdpp_files, False):
        files.append(hdp_file)
        
        hdpp_doc = parse_document(hdp_file.open('rb'))
        root = hdpp_doc.getroot()

        element_material = get_text(root.find('partdata'), 'material')
        
        xml_teeth = root.findall('teeth/tooth')
        teeth = []
        for xml_tooth in xml_teeth:
            tooth_type = xml_tooth.get('type')
            tooth_number = get_text(xml_tooth, 'tno')
            implant_system = None

            implant_system_node = xml_tooth.find('implantconnect')
            if implant_system_node is not None:
                implant_system = get_text(implant_system_node, 'system')

            tooth = Tooth(fdi=tooth_number, color=None, type=tooth_type,
                    is_anatomical=False, implant_type=None, implant_system=implant_system)
            teeth.append(tooth)

        element = Element(file_name=hdp_file.with_suffix('.stl').name, material=element_material, color=None, height=0.0,
                is_screw_retained=False, type=None, teeth=teeth)
        elements.append(element)
        
    return files, Order(format=FORMAT_NAME, project=None, patient=None, comment=None,
        practice_id=None, elements=elements, models=[], scan_files=[], loaded_scene_file=None)


import unittest

from cad_parser.definitions import Tooth

class TestTooth(unittest.TestCase):
    def test_constructor(self):
        tooth = Tooth(fdi=1, color='A1', implant_system=None, implant_type=None, is_anatomical=None, type='AnatomicCrown')

        self.assertEqual(tooth.fdi, 1)
        self.assertEqual(tooth.color, 'A1')
        self.assertEqual(tooth.implant_type, None)
        self.assertEqual(tooth.is_anatomical, None)
        self.assertEqual(tooth.type, 'AnatomicCrown')



from typing import List, Tuple

from cad_parser.errors import ExocadParserError
from cad_parser.definitions import Order, Element, Tooth, Patient, ScanFile
from cad_parser.utils import consolidate_colors, parse_document, get_text

from pathlib import Path

import logging
logger = logging.getLogger(__name__)

FORMAT_NAME = 'exocad'

def is_supported(file_path: Path) -> bool:
    '''
    Is the given file path a supported ExoCAD project?

    Parameters
    ----------
    file_path : Path
        The file path to check.

    Returns
    -------
    bool
        True if the file path is a supported ExoCAD project, False otherwise.
    '''
    return next(file_path.glob('*.dental[pP]roject'), False) and (
        next(file_path.glob('*.construction[iI]nfo'), False) or
        next(file_path.glob('*.model[iI]nfo'), False) or
        next(file_path.glob('*.scan[iI]nfo'), False) or
        next(file_path.glob('*.partial[iI]nfo'), False)
    )

def parse(project_path: Path) -> Tuple[Path, Path, Path, Path, Path, Order]:
    '''
    Parse an exocad project and return an `Order` object.

    The process of creating an exocad construction follows the following steps:
    1. Scan the patient's dental model resulting in a `.scanInfo` file.
    2. Create a model from the scanned model resulting in a `.modelInfo` file.
    3a. Either create a construction from the model resulting in a `.constructionInfo` file.
    3b. Or create a partial construction from the model resulting in a `.partialInfo` file.

    Thus, the order contains the elements defined either in the `constructionInfo` if present, or in the `partialInfo` file.

    Parameters
    ----------
    project_path : Path
        The path to the exocad project directory.

    Raises
    ------
    ExocadParseError
        If none of the required files are missing or if they cannot be parsed correctly.

    Returns
    -------
    dental_project_path : Path
        The path to the exocad project file.
    scan_info_path : Path
        The path to the `.scanInfo` file.
    model_info_path : Path
        The path to the `.modelInfo` file.
    partial_info_path : Path
        The path to the `.partialInfo` file.
    construction_info_path : Path
        The path to the `.constructionInfo` file.
    order : Order
        The order object containing the elements, models and scans defined in the exocad project.

    '''
    if not is_supported(project_path):
        raise ExocadParserError(f"Unsupported exocad project: {project_path}")
    
    project = None
    patient = None
    comment = None
    practice_id = None
    elements = []
    models = []
    scan_files = []
    loaded_scene_file = None

    # Get basic information
    if dental_project_path := next(project_path.glob('*.dentalProject'), False):
        patient = parse_dental_project(dental_project_path)

    # get infomation if the project is in the scan step
    if scan_info_path := next(project_path.glob('*.scanInfo'), False):
        scan_order = parse_scan_info(scan_info_path)
        project = scan_order.project
        comment = scan_order.comment
        practice_id = scan_order.practice_id
        scan_files = scan_order.scan_files

    # get information if the project is in the model step
    if model_info_path := next(project_path.glob('*.modelInfo'), False):
        model_order = parse_model_info(model_info_path)
        project = model_order.project
        comment = model_order.comment
        practice_id = model_order.practice_id
        models = model_order.models
        scan_files = model_order.scan_files
        loaded_scene_file = model_order.loaded_scene_file

    if partial_info_path := next(project_path.glob('*.partialInfo'), False):
        elements = parse_partial_info(partial_info_path)

    # get the information if the project is in the construction step
    if construction_info_path := next(project_path.glob('*.constructionInfo'), False):
        construction_order = parse_construction_info(construction_info_path)
        project = construction_order.project
        comment = construction_order.comment
        practice_id = construction_order.practice_id
        elements = construction_order.elements
        scan_files = construction_order.scan_files
        loaded_scene_file = construction_order.loaded_scene_file

    order = Order(
        format=FORMAT_NAME,
        project=project,
        patient=patient,
        comment=comment,
        practice_id=practice_id,
        elements=elements,
        models=models,
        scan_files=scan_files,
        loaded_scene_file=loaded_scene_file
    )

    return dental_project_path, scan_info_path, model_info_path, partial_info_path, construction_info_path, order
    
def parse_dental_project(dental_project_path: Path) -> Patient:
    '''
    Parse a dental project file and extract the necessary information.

    Parameters
    ----------
    dental_project_path : Path
        The path to the dental project file.

    Returns
    -------
    Patient
        The parsed patient information.
    '''
    dp_doc = parse_document(dental_project_path.open('rb'))

    patient_id = get_text(dp_doc, 'Patient/PatientId')
    patient_first_name = get_text(dp_doc, 'Patient/PatientFirstName')
    patient_last_name = get_text(dp_doc, 'Patient/PatientName')

    return Patient(id=patient_id, first_name=patient_first_name, last_name=patient_last_name)

def parse_construction_info(construction_info_path: Path) -> Order:
    ci_doc = parse_document(construction_info_path.open('rb'))

    project_name = get_text(ci_doc, "ProjectName")
    practice_id = get_text(ci_doc, "PracticeId")
    patient_id = get_text(ci_doc, "PatientId")
    comment = get_text(ci_doc, "Notes")

    loaded_scene_file = None
    # we need to look for the most recent changed file, as the workflow of 
    # exocad users is to save their intermediate progress in a separate file.
    recentModification = 0
    for scene_file in construction_info_path.parent.glob('*.dentalCAD'):
        modified = scene_file.stat().st_mtime
        if modified > recentModification:
            recentModification = modified
            loaded_scene_file = scene_file.name

    elements = []
    for element in ci_doc.findall("ConstructionFileList/ConstructionFile"):
        file_name = get_text(element, "Filename")
        material = get_text(element, "Material")
        color = None
        height = float(get_text(element, "Height"))
        is_screw_retained = get_text(element, "IsScrewRetained")
        part_type = get_text(element, "PartType")

        teeth = []
        tooth_numbers = [tooth.text for tooth in element.findall('ToothNumbers/int')]

        for tooth_number in tooth_numbers:
            tooth_nodes = ci_doc.xpath(f'Teeth/Tooth[contains(Number,"{tooth_number}")]')

            if len(tooth_nodes) == 0:
                raise ExocadParserError(f'Tooth #{tooth_number} missing')

            tooth_node = next(iter(tooth_nodes))

            tooth_color = get_text(tooth_node, 'ToothColor')
            tooth_type = get_text(tooth_node, 'ReconstructionType')
            tooth_is_anatomical = None
            tooth_implant_type = get_text(tooth_node, 'ImplantType')
            implant_system = None

            if is_screw_retained == 'true':
                implant_system = get_text(tooth_node, 'ImplantLibraryEntryDescriptor')

            tooth = Tooth(fdi=tooth_number, color=tooth_color, type=tooth_type,
                    is_anatomical=tooth_is_anatomical, implant_type=tooth_implant_type, implant_system=implant_system)
            teeth.append(tooth)

        color = consolidate_colors(teeth)

        element = Element(file_name=file_name, material=material, color=color, height=height,
                is_screw_retained=is_screw_retained, type=part_type, teeth=teeth)
        elements.append(element)

    scanFiles = []
    for scanFile in ci_doc.findall("ScanFiles/ScanFile"):
        part_type = get_text(scanFile, 'PartType')
        file_name = get_text(scanFile, 'FileName')

        scanFiles.append(ScanFile(file_name=file_name, type=part_type))

    patient = Patient(id=patient_id, last_name=None, first_name=None)

    return Order(format=FORMAT_NAME, project=project_name, patient=patient, comment=comment,
        practice_id=practice_id, elements=elements, models=[], scan_files=scanFiles, loaded_scene_file=loaded_scene_file)

def parse_model_info(model_info_path: Path) -> Order:
    mi_doc = parse_document(model_info_path.open('rb'))

    project_name = get_text(mi_doc, "ProjectName")
    practice_id = get_text(mi_doc, "PracticeId")
    patient_id = get_text(mi_doc, "PatientId")
    comment = get_text(mi_doc, "Notes")
    is_screw_retained = None

    models = []
    for element in mi_doc.findall("ConstructionFileList/ConstructionFile"):
        file_name = get_text(element, "Filename")
        material = get_text(element, "Material")
        color = None
        height = float(get_text(element, "Height"))
        element_type = get_text(element, 'PartType')

        teeth = []
        tooth_numbers = [tooth.text for tooth in element.findall('ToothNumbers/int')]
        for tooth_number in tooth_numbers:
            tooth_nodes = mi_doc.xpath(f'Teeth/Tooth[contains(Number,"{tooth_number}")]')

            if len(tooth_nodes) == 0:
                raise ExocadParserError(f'Tooth #{tooth_number} missing')
            
            tooth_node = next(iter(tooth_nodes))
            tooth_color = get_text(tooth_node, 'ToothColor')
            tooth_type = get_text(tooth_node, 'ReconstructionType')
            tooth_is_anatomical = None
            tooth_implant_type = None
            implant_system = None

            tooth = Tooth(fdi=tooth_number, color=tooth_color, type=tooth_type,
                    is_anatomical=tooth_is_anatomical, implant_type=tooth_implant_type, implant_system=implant_system)
            teeth.append(tooth)

        color = consolidate_colors(teeth)

        model = Element(file_name=file_name, material=material, color=color, height=height,
                is_screw_retained=is_screw_retained, type=element_type, teeth=teeth)
        models.append(model)

    scanFiles = []
    for scanFile in mi_doc.findall("ScanFiles/ScanFile"):
        part_type = get_text(scanFile, 'PartType')
        file_name = get_text(scanFile, 'FileName')

        scanFiles.append(ScanFile(file_name=file_name, type=part_type))

    patient = Patient(id=patient_id, last_name=None, first_name=None)

    return Order(format=FORMAT_NAME, project=project_name, patient=patient, comment=comment,
        practice_id=practice_id, elements=[], models=models, scan_files=scanFiles, loaded_scene_file=None)

def parse_scan_info(scan_info_path: Path) -> Order:
    si_doc = parse_document(scan_info_path.open('rb'))

    project_name = get_text(si_doc, "ProjectName")
    practice_id = get_text(si_doc, "PracticeId")
    patient_id = get_text(si_doc, "PatientId")
    comment = get_text(si_doc, "Notes")

    scanFiles = []
    for scanFile in si_doc.findall("ScanFiles/ScanFile"):
        part_type = get_text(scanFile, 'PartType')
        file_name = get_text(scanFile, 'FileName')

        scanFiles.append(ScanFile(file_name=file_name, type=part_type))

    patient = Patient(id=patient_id, last_name=None, first_name=None)

    return Order(format=FORMAT_NAME, project=project_name, patient=patient, comment=comment,
        practice_id=practice_id, elements=[], models=[], scan_files=scanFiles, loaded_scene_file=None)

def parse_partial_info(partial_info_path: Path) -> List[Element]:

    si_doc = parse_document(partial_info_path.open('rb'))

    elements = []
    for element in si_doc.findall("PartialFileList/PartialFile"):
        file_name = get_text(element, 'FileName')
        material = get_text(element, 'Material')
        element_type = None

        claps_curve = False
        for buildParameter in si_doc.findall("BuildParameterLists/BuildParameters"):
            if buildParameter.get('type') == 'ClaspCurve':
                claps_curve = True
                break

        element_type = 'modelclip' if claps_curve else 'framework'

        element = Element(file_name=file_name, material=material, color=None, height=None,
                is_screw_retained=None, teeth=[], type=element_type)
        elements.append(element)
    return elements


from typing import NamedTuple, List

# Output format
class Patient(NamedTuple):
    id: int
    last_name: str
    first_name: str

class Tooth(NamedTuple):
    fdi: int
    color: str
    implant_type: str
    implant_system: str
    is_anatomical: bool
    type: str

class Element(NamedTuple):
    file_name: str
    material: str
    color: str
    height: float
    is_screw_retained: bool
    type: str
    teeth: List[Tooth]
    
class ScanFile(NamedTuple):
    file_name: str
    type: str

class Order(NamedTuple):
    format: str
    project: str
    patient: Patient
    comment: str
    practice_id: int
    elements: List[Element]
    models: List[Element]
    scan_files: List[ScanFile]
    loaded_scene_file: str



from pathlib import Path
from typing import Tuple
import re

from cad_parser.errors import ThreeOXParserError
from cad_parser.definitions import Order, Element, Tooth, Patient, ScanFile
from cad_parser.utils import parse_document, unn_to_fdi

FORMAT_NAME = '3ox'

def is_supported(file_path: Path) -> bool:
    '''
    Is the given file path a supported ExoCAD project?

    Parameters
    ----------
    file_path : Path
        The file path to check.

    Returns
    -------
    bool
        True if the file path is a supported ExoCAD project, False otherwise.
    '''
    return next(file_path.glob('*.3ox'), False)

def parse(project_path: Path) -> Tuple[Path, Order]:
    if not is_supported(project_path):
        raise ThreeOXParserError(f"Unsupported threeshape project: {project_path}")
    
    if threeshape_path := next(project_path.glob('*.3ox'), False):
        order = parse_3ox(threeshape_path)

    return threeshape_path, order

def parse_3ox(threeshape_file: Path):
    threeox_doc = parse_document(threeshape_file.open('rb'))
    root = threeox_doc.getroot()
    nsmap = root.nsmap

    # What the ...?
    nsmap['ns'] = nsmap.get(None)
    del nsmap[None]

    # FIXME: some versions don't seem to have OrderInfo elements
    # Also check ThreeShapeOrderNo as project name instead of (sometimes missing)
    # patient names.
    order_node = root.find('ns:Command/ns:AddOrderRequest/ns:Order', namespaces=nsmap)

    if order_node is None:
        raise ThreeOXParserError("Missing element: Order")

    restoration_order_node = order_node.find('ns:InnerOrder/ns:RestorationOrder', namespaces=nsmap)

    if restoration_order_node is None:
        raise ThreeOXParserError("Missing element: RestorationOrder")

    order_info_node = restoration_order_node.find('ns:OrderInfo', namespaces=nsmap)

    if order_info_node is None:
        raise ThreeOXParserError("Missing element: OrderInfo")

    patient_first_name = None
    patient_last_name = None

    patient_last_name_node = order_info_node.find('ns:Patient_LastName', namespaces=nsmap)

    if patient_last_name_node is not None:
        patient_last_name = patient_last_name_node.text

    patient_first_name_node = order_info_node.find('ns:Patient_FirstName', namespaces=nsmap)

    if patient_first_name_node is not None:
        patient_first_name = patient_first_name_node.text

    project = '_'.join(filter(None, [patient_last_name, patient_first_name]))

    if len(project) == 0:
        ts_order_number_node = order_info_node.find('ns:ThreeShapeOrderNo', namespaces=nsmap)

        if ts_order_number_node is None:
            raise ThreeOXParserError("Can't determine project number. Patient_FirstName, Patient_LastName and ThreeShapeOrderNo missing")

        project = ts_order_number_node.text

    patient_id = None
    patient = Patient(id=patient_id, last_name=patient_last_name, first_name=patient_first_name)

    comments_node = order_node.find('ns:Comments', namespaces=nsmap)

    if comments_node is None:
        raise ThreeOXParserError("Missing element: Comments")

    comments = comments_node.text if comments_node.text else ''

    element_nodes = restoration_order_node.findall('ns:ModelElements/ns:ModelElement',
            namespaces=nsmap)

    scan_files = []
    elements = []
    models = []

    for element_node in element_nodes:
        material = None
        color = None
        height = None
        type = None

        type_node = element_node.find('ns:LinkTypeClass', namespaces=nsmap)
        if type_node is not None:
            type = type_node.text

        material_node = element_node.find("ns:Material", namespaces=nsmap)

        if material_node is not None:
            material = material_node.get("displayName")

        color_node = element_node.find("ns:Color", namespaces=nsmap)

        if color_node is not None:
            color = color_node.get("displayName")

        height_node = element_node.find('ns:ModelHeight', namespaces=nsmap)

        if height_node is not None and height_node.text is not None:
            height = float(height_node.text)

        cam_file_nodes = element_node.findall('ns:CAMFiles/ns:CAMFile', namespaces=nsmap)
        file_name = None

        implant_root = None
        for x in cam_file_nodes:
            path = x.get('path')
            
            if '.stl' in path and not re.search(r'Die_[0-9]{1,2}', path):
                file_name = path.split('\\')[-1]
            elif '.stl' in path and re.search(r'Die_[0-9]{1,2}', path):
                model_path = path.split('\\')[-1]
                model_material = None
                model_color = None
                model_height = 0

                model_fdi = unn_to_fdi(int(model_path.split('_')[-1].split('.')[0]))

                model_tooth = Tooth(fdi=model_fdi, color=None, implant_type=None, implant_system=None, type='modeldie', is_anatomical=False)

                model = Element(file_name=model_path, material=model_material, color=model_color, height=model_height, type='modeldie', teeth=[model_tooth], is_screw_retained=False)
                models.append(model)
            elif '.xml' in path and 'ImplantDirectionPosition' in path:
                implant_file_name = path.split('\\')[-1]
                implant_file = parse_document(threeshape_file.with_name(implant_file_name).open('rb'))
                implant_root = implant_file.getroot()

        # if we couldn't identify a file name for the element, we'll skip it
        if not file_name:
            continue

        teeth = []

        for tooth_node in element_node.findall('ns:TypeIDs/ns:TypeID', namespaces=nsmap):
            # TODO Verify there is no material in teeth
            unn = tooth_node.get('uNN')
            fdi = unn_to_fdi(int(unn))
            tooth_color = None

            tooth_type = tooth_node.get('typeClass')
            implant_type = tooth_node.get('abutmentKitID')
            implant_system = None

            if implant_root is not None:
                implant_system_node = implant_root.find(f'ToothIndex[@unn="{unn}"]/Kit/Name')
                if implant_system_node is not None:
                    implant_system = implant_system_node.get('value')

            tooth = Tooth(fdi=fdi, color=tooth_color, type=tooth_type, is_anatomical=None,
                    implant_type=implant_type, implant_system=implant_system)
            teeth.append(tooth)

        for scan_node in element_node.findall('ns:ScanFiles/ns:ScanFile', namespaces=nsmap):
            scan_file_name = scan_node.get('path').split('\\')[-1]
            scan_type = scan_node.get('fileType')
            scan_file = ScanFile(file_name=scan_file_name, type=scan_type)
            if scan_file not in scan_files:
                scan_files.append(scan_file)

        # is_screw_retained
        element = Element(file_name=file_name, type=type, material=material, color=color, height=height,
                is_screw_retained=None, teeth=teeth)
        elements.append(element)

    # patient_id
    patient = Patient(id=None, last_name=patient_last_name, first_name=patient_first_name)

    # unavailable information
    # practice_id
    return Order(format=FORMAT_NAME, project=project, patient=patient, comment=comments,
        practice_id=None, elements=elements, models=models, scan_files=scan_files, loaded_scene_file=None)



import unittest

from cad_parser import threeshape

from pathlib import Path

class SkipTest:
    def __init__(self, condition, description):
        self.condition = condition
        self.description = description

    def __call__(self, func):
        if not self.condition:
            return unittest.skip(self.description)(func)
        return func

@SkipTest(Path('tests/samples/3shape').exists(), "Lokal tests only")
class Threeshape(unittest.TestCase):
    def test_kappe(self):
        filePath = Path('tests/samples/3shape/01_Kappe/24-07-23_Kappen_3ox/24-07-23_Kappen_3ox.3ox')
        order = threeshape.parse_3ox(filePath)

        self.assertEqual(order.format, '3ox')
        self.assertEqual(order.project, '24-07-23_Kappen_3ox')
        self.assertEqual(order.comment, '')
        self.assertEqual(len(order.elements), 5)
        for idx,element in enumerate(order.elements):
            if idx == 0:
                self.assertEqual(element.file_name, '24-07-23_Kappen_3ox_0.stl')
            if idx == 1:
                self.assertEqual(element.file_name, '24-07-23_Kappen_3ox_1.stl')
            if idx == 2:
                self.assertEqual(element.file_name, '24-07-23_Kappen_3ox_2.stl')
            if idx == 3:
                self.assertEqual(element.file_name, '24-07-23_Kappen_3ox_3.stl')
            if idx == 4:
                self.assertEqual(element.file_name, '24-07-23_Kappen_3ox_4.stl')

        self.assertEqual(len(order.scan_files), 2)
        for scan_file in order.scan_files:
            if scan_file.type == 'Preparation':
                self.assertEqual(scan_file.file_name, 'PreparationScan.dcm')
            if scan_file.type == 'AntagonistModel':
                self.assertEqual(scan_file.file_name, 'AntagonistScan.dcm')

    def test_parse_die(self):
        filePath = Path('tests/samples/3shape/01_Kappe/24-07-23_Kappen_3ox/24-07-23_Kappen_3ox.3ox')
        order = threeshape.parse_3ox(filePath)

        for idx, model in enumerate(order.models):
            if idx == 0:
                self.assertEqual(model.file_name, '24-07-23_Kappen_3ox_0_Die_9.stl')
            if idx == 1:
                self.assertEqual(model.file_name, '24-07-23_Kappen_3ox_1_Die_11.stl')
            if idx == 2:
                self.assertEqual(model.file_name, '24-07-23_Kappen_3ox_1_Die_8.stl')
            if idx == 3:
                self.assertEqual(model.file_name, '24-07-23_Kappen_3ox_2_Die_7.stl')

    def test_brueke(self):
        filePath = Path('tests/samples/3shape/08_Sekundärteil_Brücke/single_3shape_154202/Tech_01_Bork_Gruenter_20230614_0900.3ox')
        order = threeshape.parse_3ox(filePath)

        self.assertEqual(order.format, '3ox')
        self.assertEqual(order.project, 'Bork_Gruenter')
        self.assertEqual(order.comment, '')
        self.assertEqual(len(order.elements), 2)
        for idx,element in enumerate(order.elements):
            if idx == 0:
                self.assertEqual(element.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_0.stl')
            if idx == 1:
                self.assertEqual(element.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_1.stl')

        self.assertEqual(len(order.scan_files), 1)
        for scan in order.scan_files:
            if scan.type == 'Preparation':
                self.assertEqual(scan.file_name, 'PreparationScan.dcm')

        for idx, model in enumerate(order.models):
            if idx == 0:
                self.assertEqual(model.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_0_Die_22.stl')
            if idx == 1:
                self.assertEqual(model.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_1_Die_27.stl')

    def test_is_supported(self):
        # Positive test cases
        self.assertTrue(threeshape.is_supported(Path('tests/samples/3shape/01_Kappe/24-07-23_Kappen_3ox')))
        self.assertTrue(threeshape.is_supported(Path('tests/samples/3shape/08_Sekundärteil_Brücke/single_3shape_154202')))
        self.assertTrue(threeshape.is_supported(Path('tests/samples/3shape/08_Verblendete_Brücke/179439')))

        # Negative test cases
        self.assertFalse(threeshape.is_supported(Path('tests/samples/exocad/01_Kappe/Menne_Kappe')))
        self.assertFalse(threeshape.is_supported(Path('tests/samples/exocad/02_Krone/156094_krone_inkl_Abutment')))

    def test_parse(self):
        threeshape_file, order = threeshape.parse(Path('tests/samples/3shape/08_Sekundärteil_Brücke/single_3shape_154202'))

        self.assertEqual(threeshape_file, Path('tests/samples/3shape/08_Sekundärteil_Brücke/single_3shape_154202/Tech_01_Bork_Gruenter_20230614_0900.3ox'))

        self.assertEqual(order.format, '3ox')
        self.assertEqual(order.project, 'Bork_Gruenter')
        self.assertEqual(order.comment, '')
        self.assertEqual(len(order.elements), 2)
        for idx,element in enumerate(order.elements):
            if idx == 0:
                self.assertEqual(element.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_0.stl')
            if idx == 1:
                self.assertEqual(element.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_1.stl')

        self.assertEqual(len(order.scan_files), 1)
        for scan in order.scan_files:
            if scan.type == 'Preparation':
                self.assertEqual(scan.file_name, 'PreparationScan.dcm')

        for idx, model in enumerate(order.models):
            if idx == 0:
                self.assertEqual(model.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_0_Die_22.stl')
            if idx == 1:
                self.assertEqual(model.file_name, 'Tech_01_Bork_Gruenter_20230614_0900_1_Die_27.stl')

    def test_element_files_exist(self):
        threeshape_file, order = threeshape.parse(Path('tests/samples/3shape/02_Krone/180148'))

        for element in order.elements:
            self.assertIsNotNone(element.file_name)

    def test_element_type_can_be_none(self):
        threeshape_file, order = threeshape.parse(Path('tests/samples/3shape/03_Teleskop/164182'))

        for element in order.elements:
            self.assertIsNotNone(element.type)

        threeshape_file, order = threeshape.parse(Path('tests/samples/3shape/02_Krone/180148'))

        for element in order.elements:
            self.assertIsNone(element.type)

    def test_abutment_alias(self):
        threeshape_file, order = threeshape.parse(Path('tests/samples/3shape/09_Abutment/180066'))

        for element in order.elements:
            for tooth in element.teeth:
                if tooth.fdi == 46:
                    self.assertEqual(tooth.implant_system, 'Abutment_MEG_3.555_GH05')
                if tooth.fdi == 45:
                    self.assertEqual(tooth.implant_system, 'Abutment_MEG_3.555_GH05')
                if tooth.fdi == 14:
                    self.assertEqual(tooth.implant_system, 'Abutment_MEG_3.555_GH05')
                if tooth.fdi == 17:
                    self.assertEqual(tooth.implant_system, 'Abutment_MEG_3.555_GH05')

        
        threeshape_file, order = threeshape.parse(Path('tests/samples/3shape/09_Abutment/180608'))

        for element in order.elements:
            for tooth in element.teeth:
                if tooth.fdi == 14:
                    self.assertEqual(tooth.implant_system, 'Abutment_MEG_3.555_GH05')
                if tooth.fdi == 36:
                    self.assertEqual(tooth.implant_system, 'Abutment_MEG_3.555_GH05')

        threeshape_file, order = threeshape.parse(Path('tests/samples/3shape/09_Abutment/184819'))

        for element in order.elements:
            for tooth in element.teeth:
                if tooth.fdi == 16:
                    self.assertEqual(tooth.implant_system, 'BTI® 23.2 INTERNA UNIVERSAL_INPSPU0_65')
                if tooth.fdi == 13:
                    self.assertEqual(tooth.implant_system, 'BTI® 23.2 INTERNA UNIVERSAL_INPSPU0_65')


import unittest

from pathlib import Path

from cad_parser.hyperdent import is_supported, parse

class SkipTest:
    def __init__(self, condition, description):
        self.condition = condition
        self.description = description

    def __call__(self, func):
        if not self.condition:
            return unittest.skip(self.description)(func)
        return func

@SkipTest(Path('tests/samples/hyperdent').exists(), "Lokal tests only")
class TestHyperdent(unittest.TestCase):
    def test_is_supported(self):
        self.assertTrue(is_supported(Path('tests/samples/hyperdent/09_abutment/180066')))

    def test_parse(self):
        files, order = parse(Path('tests/samples/hyperdent/09_abutment/180066'))
        
        for element in order.elements:
            for tooth in element.teeth:
                if tooth.type == 'abutment':
                    self.assertEqual(tooth.implant_system, 'Abutment_MEG_3.555_GH05')

    def test_filenames(self):
        files, order = parse(Path('tests/samples/hyperdent/09_abutment/180066'))

        for element in order.elements:
            segments = element.file_name.split('/')
            self.assertEqual(len(segments), 1)
            segments = element.file_name.split(r'\\')
            self.assertEqual(len(segments), 1)
            self.assertIn('.stl', element.file_name)